/*
 *
 *
 *    AutoCap_ArduinoUno_v1.5.ino
 *
 *    AutoCap Firmware for Arduino
 *    Version 1.5
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

//
// See Protocol.txt for a description of the data protocol used by this firmware.
//

// local includes
#include "Interfaces.h"

// normal DC motor support (H-bridge, up/down polarity/direction, etc.)
#include "MotorDC.h"

///////////////////////////////////////////////////////////////////////////
//
//                    START CONFIGURATION SETTINGS                   
//
//  This section defines several items that you can modify to match the
//  firmware to your particular hardware.  The default configuration is:
//     + Motor interface is an Arduino Motor Shield R3.
//     + Two simple DC motor ports on the motor shield.
//     + Default serial port (USB port for Arduinos that support it).
//     + 9600 baud, no parity, 8-bit data, 1 stop bit.
//     + User controls (e.g., buttons and displays) are DISABLED.
//     + Stepper motor support is DISABLED.
//
//  If you have something other than the defaults shown above, this is
//  the place to change the configuration to match your needs.
//
//  CAUTION: Enable only the features that you need for your hardware.
//           The smaller Arduino boards have very little memory (RAM),
//           and there are enough features available to overrun the
//           available memory, even on an UNO R3.  Start with only the
//           features that you need, like motor definitions, and then
//           slowly add features from there.  If your controller starts
//           exhibiting strange or random behavior, it is most likely
//           because you have enabled too many features, or configured
//           too many motor objects for the amount of available memory.
//           Reducing the number of features and/or the number of motors
//           will help you fit the firmware on the board you have.
//
//           If you find that your device does not have enough memory
//           for the features you want, try moving to a different board.
//           The Arduino Mega 2560 R3 boards have FOUR TIMES the available
//           RAM of the UNO R3.  The Mega 2560 boards also have more I/O
//           pins for things like motors, LEDs, and user buttons, as well
//           as multiple dedicated UARTs (serial ports) which can be used
//           to drive displays like the serial backpack.
//
//  To modify this sketch, you will need to have a basic understanding of
//  the Arduino software environment, and some simple language constructs
//  such as comments.  The Arduino team offers an introduction on their
//  website:
//
//     http://arduino.cc/en/Guide/HomePage
//
///////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: stepper motor support library
//
//  Stepper motor support requires the AccelStepper library, available from:
//
//     http://www.airspayce.com/mikem/arduino/AccelStepper/classAccelStepper.html
//
//  If you have installed the AccelStepper library, then you can set
//  USE_ACCELSTEPPER to 'true' to enable stepper motor support.  You
//  can then define one or more stepper motor objects.
//
//  If you do not intend to use stepper motors, leave this set to 'false'.
//
#define USE_ACCELSTEPPER false
//
//  ...a 'true' value will cause the following headers to be loaded...
//
#if USE_ACCELSTEPPER == true
	// stepper motor support based on AccelStepper library (don't change these)
	#include <AccelStepper.h>
	#include "MotorAccel.h"
#endif


///////////////////////////////////////////////////////////////////////////
//
//  REQUIRED: Motor Objects - Define one or more motor control ports.
//
//  For each physical motor port, you must provide one motor object here.
//  The defaults shown below are for a DC motor interface using the Arduino
//  Motor Shield R3, which has two independent motor ports for connection
//  to normally commutated DC motors.
//
//  Each motor object will have a type (e.g., MotorDC or MotorAccel), and a
//  name (e.g., MOTOR1).  The motor objects also take a number of additional
//  arguments, most of which are Arduino pin numbers.  These arguments are
//  placed between the parenthesis next to the motor name, as shown in the
//  examples below.
//
//  See the README.txt file for more information and examples regarding
//  motor port configuration.
//

//
// EXAMPLE: Create a new DC motor port; set any pin to (-1) to disable feature.
//          This port matches the first port of the Arduino Motor Shield R3.
//          See MotorDC.h for pin assignment details for DC motors.
//
MotorDC MOTOR1(
	3,		// PWM/ENABLE output pin - MUST be a PWM-capable pin
	12,		// DIRECTION output pin
	(-1),	// UP/IN1 output pin
	(-1),	// DOWN/IN2 output pin
	0,		// SENSE input A/D pin (0 = A0 pin)
	(-1),	// LED output pin
	9,		// BRAKE output pin
	true	// BRAKE PWM Enable (must be true or false)
);

//
// EXAMPLE: Create a second DC motor port, with different pin assignments.
//          This port matches the second port of the Arduino Motor Shield R3.
//          See MotorDC.h for pin assignment details for DC motors.
//
MotorDC MOTOR2(
	11,		// PWM/ENABLE output pin - MUST be a PWM-capable pin
	13,		// DIRECTION output pin
	(-1),	// UP/IN1 output pin
	(-1),	// DOWN/IN2 output pin
	1,		// SENSE input A/D pin (1 = A1 pin)
	(-1),	// LED output pin
	8,		// BRAKE output pin
	true	// BRAKE PWM Enable (must be true or false)
);

//
// EXAMPLE: Create a single outboard H-bridge for a DC motor; with EN = D3, IN = { D4, D5 }
//
/*
MotorDC MOTOR1(
	3,		// PWM/ENABLE output pin - MUST be a PWM-capable pin
	(-1),	// DIRECTION output pin
	4,		// UP/IN1 output pin
	5,		// DOWN/IN2 output pin
	(-1),	// SENSE input A/D pin (0 = A0 pin)
	(-1),	// LED output pin
	(-1),	// BRAKE output pin
	false	// BRAKE PWM Enable (must be true or false)
);
*/

//
// EXAMPLE: Create a single dual H-bridge for a DC motor.  This is the example
//          from the website article: http://www.kk5jy.net/AutoCap-v1/Firmware/
//
/*
EnablePin enablePin(
	3,		// pin D3
	true,	// pin is PWM
	0xB0,	// PWM when running
	true,	// disable-on-stop
	0x40);	// PWM when stopped

MotorAccel MOTOR1(
	AccelStepper::FULL4WIRE,
	4,		// IN1 = D4
	5,		// IN2 = D5
	A4,		// IN3 = A4
	A5,		// IN4 = A5
	0xFF,	// no LED
	400.0,	// max speed = 400 steps/sec
	&enablePin, // enable pin definition (see above)
	1,		// one enable pin
	0xFF);	// brake delay = 255msec
*/


///////////////////////////////////////////////////////////////////////////
//
//  REQUIRED: list of motor objects
//
//  Each motor object defined above should be listed here, with each motor
//  name prefixed by an ampersand (&), and the list separated by commas.
//  The two motors listed below match the two example MotorDC example ports
//  shown above.  The order in which they are listed is the order in which
//  port numbers will be assigned; the first one is port zero, the second
//  one is port one, and so on.
//
//  In programmer-speak, this is an "array of IMotor object pointers".  That
//  is why each name must start with an ampersand.  This array is where the
//  firmware looks whenever a specific port number is requested.
//  
static IMotor * const motors[] = {
	& MOTOR1,
	& MOTOR2
};


///////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: User Control Pin Definitions
//
//  If you have any user controls, you will need to define the pins used by
//  your buttons, etc., below.  Set any pin to (0) or (-1) to disable the
//  associated feature.  By default, all of these are disabled.
//
//  MOTOR_UP    - When this pin is brought and held LOW, the motor will be
//                continuously moved forward at 100% effort (or proportionally
//                to the voltage provided to the SPEED pin) until the pin
//                is returned HIGH.
//  MOTOR_DN    - When this pin is brought and held LOW, the motor will be
//                continuously moved backward at 100% effort (or proportionally
//                to the voltage provided to the SPEED pin) until the pin
//                is returned HIGH.
//  MOTOR_STOP  - When this pin is brought momentarily LOW, any current motor
//                movement command is canceled, and the motor stopped.  This
//                can be used to cancel a host command before it completes.
//  PORT_SELECT - When this pin is brought momentarily LOW, the motor port
//                affected by UP, DOWN, STOP, and SPEED is incremented by one.
//  USER_MODE   - Selects the user display/config/operation mode.
//
#define MOTOR_UP (-1)
#define MOTOR_DN (-1)
#define MOTOR_STOP (-1)
#define PORT_SELECT (-1)
#define USER_MODE (-1)


///////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: Persist settings to EEPROM
//
//  Setting this symbol to true will cause certain settings to be persistet
//  to EEPROM whenever they are changed, and read from EEPROM upon startup.
//
//  This is currently only used for stepper PWM and brake settings.
//
#define PERSIST_SETTINGS false


///////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: set HEARTBEAT_LED to something other than zero to enable an
//            LED on that pin number, to blink now and then to show that
//            the program is still running, and hasn't crashed due to RFI, etc.
//
//            Using a value of zero (0) will disable this feature.
//
//            Using a value of (13) will use the LED on the UNO board.
//
//            NOTE: If you are using an Arduino Motor Shield R3, or a
//            SparkFun ArduMoto, set this value to something other than 13,
//            because pin 13 is used for the motor 'B' direction control on
//            those shields.
//
#define HEARTBEAT_LED (0)


///////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: Select host serial port
//
//  Assign a specific hardware port to be the interface to the host;
//  by default, the USB connection on most boards is 'Serial', which is
//  also the hardware TTL serial port on the UNO board.  Make sure to
//  select only ONE option.
//

// DEFAULT for most boards' USB port - IF UNSURE, KEEP THIS VALUE
#define HostSerial Serial

// hardware UART for Leonardo, and second hardware UART for MEGA board
//#define HostSerial Serial1

// third MEGA hardware UART
//#define HostSerial Serial2

// fourth MEGA hardware UART
//#define HostSerial Serial3

///////////////////////////////////////////////////////////////////////////
//
//  REQUIRED: Serial Port Speed - the 'baud rate' of the serial port
//
#define SerialRate (9600)


///////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: Enable Display Support
//
//  If you add an LED/LCD display, select the display type below and
//  configure the pins, etc.
//

//
//   Set this to 'true' to enable 16x2 display; otherwise leave 'false'
//
#define ENABLE_DISPLAY false

#if ENABLE_DISPLAY == true
//
//   Configure display I/O pins
//
// For details, see: https://www.arduino.cc/en/Reference/LiquidCrystal
//    and then set the pin numbers below to match your hardware; you
//    can use types other than 'LiquidCrystal' as long as the display
//    object is called 'display' and implements the same interface as
//    the 'LiquidCrystal' type.
//

//
//  LCD example
//
// This example is for a 16x2 LCD display connected to D6 through D12
//#include <LiquidCrystal.h>
//LiquidCrystal display(6, 7, 8, 9, 10, 11, 12);

//
//  OLED example
//
// This example uses the AdaFruit enhanced Adafruit_CharacterOLED type
// that is optimized for OLEDs. This type works with both AdaFruit and
// SparkFun displays based on the RS0010 chip.
//
// See this hookup guide for an example:
//    https://learn.sparkfun.com/tutorials/oled-display-hookup-guide/all
#include <Adafruit_CharacterOLED.h> 
Adafruit_CharacterOLED display(OLED_V2, 6, 7, 8, 9, 10, 11, 12);

//
//   Configure display size (default is 16x2)
// 
#define DISPLAY_ROWS (2)
#define DISPLAY_COLS (16)

//
//   Backlight - set this to 'true' if your display supports a
//               'setBacklight' option and you want to enable
//               the backlight; otherwise leave this set to 'false'
//
#define DISPLAY_BACKLIGHT false
#endif


///////////////////////////////////////////////////////////////////////////
//
//                    END CONFIGURATION SETTINGS                   
//
//  It is NOT recommended that you change anything beyond this point unless
//  you completely understand what your changes will do to the program.
//
///////////////////////////////////////////////////////////////////////////

// Define DEBUG to get extra debugging output statements;
//   NOTE: this extra output slows down the poor Arduino a LOT, use
//         only for diagnostic purposes, and not for normal operation
//#define DEBUG

#ifdef DEBUG
// define this to echo serial chars back to the host (for debugging only)
//#define ECHO
#endif

// include de-bounce library
#include "Bounce2.h"

// include HB macros
#include "Heartbeat.h"

//
//  Constants - things you can change, but probably shouldn't
//

// the version strings
#define VERSION_NUMBER "1.5"
#define VERSION_DATE "20160917c"
#define VERSION_TYPE "beta"
#define VERSION_STRING "#info,AutoCap for Arduino " VERSION_NUMBER " (" VERSION_TYPE "-" VERSION_DATE ")"

// limits, etc.
const int MaxLine = 16;			// maximum length of input string
const int BufferLength = 64;	// maximum length of the buffer string
const int StatusInterval = 250;	// the status report interval (msec)
const int SpeedInterval = 100;  // the user speed read interval (msec)
const char Delimiter = '\r';	// the line delimiter (from host)
const int AnalogMax = 1023;		// the highest A/D reading (10-bits for ATMega)
const int MaxPWM = 255;			// the highest PWM output (8-bits for ATMega)
const int ScaleAnalogPWM = AnalogMax / MaxPWM;	// divisor for converting
												// analog input to PWM output
const uint8_t NUM_PORTS = sizeof(motors) / sizeof(IMotor*);	// total port count
const unsigned long MinStateUpdatePeriod = 10 * 1000; // 10 seconds - min time between EEPROM updates
const uint16_t SetHoldTime = 1.1 * 1000; // must hold MODE for 2s to SET

//
//  Globals
//
String rxBuffer = "";			// a string to hold incoming data
boolean cmdReady = false;		// whether the string is complete
boolean showStatus = false;		// indicates a fresh current (I) reading
char ioBuffer[BufferLength];	// used for numeric parsing
unsigned long lastStatus = 0UL;	// timestamp of last status update
unsigned long lastSpeed = 0UL;	// timestamp of last speed update
unsigned long now = 0UL;		// global clock, amortized to one read/loop()

// the user buttons
#if MOTOR_UP != (-1)
Bounce buttonUp;
#endif
#if MOTOR_DN != (-1)
Bounce buttonDown;
#endif
#if PORT_SELECT != (-1)
Bounce buttonSel;
#endif
#if MOTOR_STOP != (-1)
Bounce buttonStop;
#endif
#if USER_MODE != (-1)
Bounce buttonMode;
unsigned long modeKeyDown = 0UL;
#endif

// set up the UI state management
#include "State.h"
State state(motors, NUM_PORTS);

#if PERSIST_SETTINGS == true
// include EEPROM-related utilities
#include "EEPROM.h"
#include "persist.h"
#if ENABLE_DISPLAY == true
unsigned long lastStateChange = 0UL;// timestamp of last status change by user
bool stateWriteNeeded = false;		// flag showing an EEPROM status write is needed
#endif // ENABLE_DISPLAY == true (display support)
#endif // PERSIST_SETTINGS == true (EEPROM support)

#if ENABLE_DISPLAY == true
//
//  Update attached displays
//
inline void UpdateDisplays(const State &state, IMotor * const motors[], bool force = false) {
	// copy of state object, used to detect state changes
	static State check(state);

	// used for forming format strings for display; perhaps the code should be refactored
	//    to eliminate snprintf()s, so that we can give back these bytes (TODO)
	// make the buffer size long enough to hold 1.5 lines of text
	static char printfBuffer[(int)(DISPLAY_COLS * 1.5)];

	bool stateChanged = (state != check);
	if (force || stateChanged) {
		//
		//  OK, so here's the deal...
		//
		//  This is where you update what goes in your display.  The state and motor
		//  objects are passed in above, so anything you can read from either of these
		//  objects can be added to your display.
		//
		//  A standard example is provided below, and matches what is demonstrated on
		//  the AutoCap website.  You can use it as-is, or make your own layout, but
		//  if you decide to change the layout, you need to do so below.
		//

		// line 1
		display.setCursor(0, 0);
		snprintf(ioBuffer, sizeof(ioBuffer), strcpy_P(printfBuffer, PSTR("Power: %3d%% ")), (int)(state.UserPWM));
		display.print(ioBuffer);
		if (motors[state.UserPort]->IsRunning()) {
			display.print(strcpy_P(ioBuffer, PSTR(" RUN")));
		} else {
			display.print(strcpy_P(ioBuffer, PSTR("    ")));
		}

		// line 2
		display.setCursor(0, 1);
		switch (state.InputMode) {
			case Momentary:
				display.print(strcpy_P(ioBuffer, PSTR("Normal          ")));
				break;
			case Continuous:
				display.print(strcpy_P(ioBuffer, PSTR("Latched         ")));
				break;
			case DoPulse:
				snprintf(ioBuffer, sizeof(ioBuffer), strcpy_P(printfBuffer, PSTR("Pulse: %4d ms  ")), (int)(state.Width));
				display.print(ioBuffer);
				break;
			case SetPulse:
				snprintf(ioBuffer, sizeof(ioBuffer), strcpy_P(printfBuffer, PSTR("Pulse: %4d ms *")), (int)(state.Width));
				display.print(ioBuffer);
				break;

#if USE_ACCELSTEPPER == true
			case DoStep:
				snprintf(ioBuffer, sizeof(ioBuffer), strcpy_P(printfBuffer, PSTR("Step : %3d st   ")), (int)(state.Steps));
				display.print(ioBuffer);
				break;
			case SetSteps:
				snprintf(ioBuffer, sizeof(ioBuffer), strcpy_P(printfBuffer, PSTR("Step : %3d st * ")), (int)(state.Steps));
				display.print(ioBuffer);
				break;
#endif
			case SetPWM:
				display.print(strcpy_P(ioBuffer, PSTR("Set Power *     ")));
				break;
			default:
				display.print(strcpy_P(ioBuffer, PSTR("???????         ")));
				break;
		}
#if PERSIST_SETTINGS == true
		if (stateChanged) {
			stateWriteNeeded = true;
			lastStateChange = now;
		}
#endif
		check = state;
	}
}
#endif


//
//  setup() - runs when the unit is booted or restarted.
//
void setup() {
	// initialize serial
	HostSerial.begin(SerialRate);

	// reserve bytes for the rxBuffer
	rxBuffer.reserve(MaxLine + 2);

	// print introductory line
	HostSerial.println(strcpy_P(ioBuffer, PSTR(VERSION_STRING)));

	#if ENABLE_DISPLAY == true
	display.begin(DISPLAY_COLS, DISPLAY_ROWS);
	delay(1000);
	display.clear();
	display.home();
	display.clear();
	display.home();
	const int x = (DISPLAY_COLS - 8) / 2; // center the banner
	display.setCursor(x, 0);
	#if DISPLAY_BACKLIGHT == true
	display.setBacklight(HIGH);
	#endif
	display.print(strcpy_P(ioBuffer, PSTR("AutoCap")));
	display.setCursor(x - 1, 1);
	display.print(strcpy_P(ioBuffer, PSTR(VERSION_DATE)));
	delay(1500);
	#endif
	
	// initialize user pins
	#if MOTOR_UP != (-1)
	buttonUp.attach(MOTOR_UP);
	#endif
	#if MOTOR_DN != (-1)
	buttonDown.attach(MOTOR_DN);
	#endif
	#if MOTOR_STOP != (-1)
	buttonStop.attach(MOTOR_STOP);
	#endif
	#if PORT_SELECT != (-1)
	buttonSel.attach(PORT_SELECT);
	#endif
	#if USER_MODE != (-1)
	buttonMode.attach(USER_MODE);
	#endif

	//
	//  TODO: If motors need extra initialization prior to starting,
	//        provide it here; this includes PWM divisors, etc.
	//
	//        Alternatively, the motor classes can be subclassed, and
	//        can override Setup() to provide extra initialization.
	//

	// configure PWM frequencies
	#if defined(__AVR_ATmega2560__)
		//
		// PWM: Change PWM frequency of Mega R3's digital output lines D6, D7, D8
		//      to ~4kHz (default is ~500Hz), as per:
		//
		//      http://sobisource.com/arduino-mega-pwm-pin-and-frequency-timer-control/
		//
		TCCR4B = (TCCR4B & 0xF8) | 0x02;
		// emit compiler warning
		#warning Using 4kHz PWM for D6, D7, D8, for Mega R3
	#endif
	#if defined(__AVR_ATmega328P__)
		//
		// PWM: Change PWM frequency of UNO R3's digital output lines D3 and D11
		//      to ~4kHz (default is ~500Hz), as per:
		//
		//      http://playground.arduino.cc/Main/TimerPWMCheatsheet
		//
		//      This changes the PWM frequency of the Motor Shield R3, as well.

		TCCR2B = (TCCR2B & 0xF8) | 0x02;
		// emit compiler warning
		#warning Using 4kHz PWM for D3, D11, for UNO R3
	#endif

	// start each motor port
	for (int i = 0; i != NUM_PORTS; ++i) {
		motors[i]->Setup();
		motors[i]->LED(i == state.UserPort);
	}

	// setup the heartbeat LED
	HeartbeatSetup();

	#if PERSIST_SETTINGS == true
	load_eeprom(true);
	#endif

	#if ENABLE_DISPLAY == true
	// display updates
	display.clear();
	UpdateDisplays(state, motors, true);
	#endif
}


//
//  loop() - the main program loop
//
void loop() {
	//
	//  run each motor object's state tracking
	//
	for (int i = 0; i != NUM_PORTS; ++i) {
		motors[i]->Run();
	}

	// read the clock
	now = millis();

	// blink the HB LED
	HeartbeatLoop();

	//
	//  STATUS - read the analog inputs measuring motor current
	//
	if (showStatus && (Elapsed(now, lastStatus) >= StatusInterval)) {
		HostSerial.print(strcpy_P(ioBuffer, PSTR("#status")));
		for(int i = 0; i != NUM_PORTS; ++i) {
			HostSerial.print(strcpy_P(ioBuffer, PSTR(",I")));
			HostSerial.print(i);
			HostSerial.print('=');
			HostSerial.print(motors[i]->ReadCurrent());
		}
		HostSerial.println();
		lastStatus = now;
	}

	//
	//  DISPLAY - update the display
	//
	#if ENABLE_DISPLAY == true
	if (state.UpdateDisplays(now)) {
		// perform the update
		UpdateDisplays(state, motors, true);
	}

	#if PERSIST_SETTINGS == true
	// do EEPROM update of new settings
	if (stateWriteNeeded
		&& (Elapsed(now, lastStateChange) >= MinStateUpdatePeriod)) {
		#ifdef DEBUG
		unsigned long start = millis();
		#endif
		
		// update the state block in EEPROM
		update_state_in_eeprom();

		#ifdef DEBUG
		HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,Update state block in eeprom for user-changed settings: ")));
		HostSerial.print(millis() - start);
		HostSerial.println(strcpy_P(ioBuffer, PSTR("ms")));
		#endif

		// clear flag
		stateWriteNeeded = false;
	}
	#endif // PERSIST_SETTINGS == true
	#endif // ENABLE_DISPLAY == true
	
	//
	//
	//  USER COMMANDS - handle button presses
	//
	//

	//
	//  UP
	//
	#if MOTOR_UP != (-1)
	if (buttonUp.update()) {
		char newState = buttonUp.read();
		state.Up(newState == LOW);
		#if ENABLE_DISPLAY == true
		// display updates
		UpdateDisplays(state, motors);
		#endif
	}
	#endif

	//
	//  DOWN
	//
	#if MOTOR_DN != (-1)
	if (buttonDown.update()) {
		char newState = buttonDown.read();
		state.Down(newState == LOW);
		#if ENABLE_DISPLAY == true
		// display updates
		UpdateDisplays(state, motors);
		#endif
	}
	#endif

	//
	//  STOP
	//
	#if MOTOR_STOP != (-1)
	if (buttonStop.update()) {
		char newState = buttonStop.read();
		state.Stop(newState == LOW);
		#if ENABLE_DISPLAY == true
		// display updates
		UpdateDisplays(state, motors);
		#endif
	}
	#endif

	//
	//  MODE
	//
	#if USER_MODE != (-1)
	if (buttonMode.update()) {
		// read the debounced state
		char newState = buttonMode.read();

		// button down?
		if (newState == LOW) {
			// start the key-down timer
			modeKeyDown = now;
			// don't allow timer value of zero, because we use that value to indicate 'timer stopped'
			if (modeKeyDown == 0)
				modeKeyDown = 1;
		} else { // button up?
			// if the key-down timer is still running, move to next mode
			if (modeKeyDown != 0) {
				state.Mode();
				modeKeyDown = 0; // cancel the key-down timer
			}
		}

		#if ENABLE_DISPLAY == true
		// display updates
		UpdateDisplays(state, motors);
		#endif // ENABLE_DISPLAY == 1
	} // USER_MODE != -1

	// process SET (holding the MODE button)
	if ((modeKeyDown != 0) && (Elapsed(now, modeKeyDown) > SetHoldTime)) {
		// toggle modes
		state.ModeHold();
		// cancel the key-down timer
		modeKeyDown = 0;

		#if ENABLE_DISPLAY == true
		// display updates
		UpdateDisplays(state, motors);
		#endif // ENABLE_DISPLAY == 1
	}
	#endif // USER_MODE != -1

	//
	//  NEXT PORT
	//
	#if PORT_SELECT != (-1)
	if (buttonSel.update() && buttonSel.read() == LOW) {
		state.PortSelect();
		#if ENABLE_DISPLAY == true
		// display updates
		UpdateDisplays(state, motors);
		#endif
	}
	#endif // PORT_SELECT != -1

	//
	//  HOST COMMANDS - if we have a complete input line from the host, process it
	//
	if (cmdReady) {
		char cmdChar = 0;

		// measure the input buffer
		const int len = rxBuffer.length();

		// need at least one character
		if (len < 1) {
			//HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Empty command")));
			goto finally;
		}

		// parse and process the string
		cmdChar = tolower(rxBuffer.charAt(0));
		switch(cmdChar) {
			//
			//  INFO
			//
			case 'i': {
				HostSerial.println(strcpy_P(ioBuffer, PSTR(VERSION_STRING)));
			} break;
			
			//
			//  COUNT
			//
			case 'c': {
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#count,")));
				HostSerial.println(NUM_PORTS);
			} break;

			//
			//  ALL-STOP
			//
			case 'z': {
				// set all PWM ports to STOP
				for (uint8_t i = 0; i != NUM_PORTS; ++i) {
					motors[i]->Stop();
				}
			} break;

			//
			//  STATUS CONTROL
			//
			case 's': {
				if (len != 2) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed STATUS command - length != 2")));
					goto finally;
				}

				// the status control, must be zero (disable) or one (enable)
				uint8_t state = rxBuffer.charAt(1) - '0';
				if (state != 0 && state != 1) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed STATUS command - Invalid state")));
					goto finally;
				}

				// set the flag; applies to both ports
				showStatus = state ? true : false;
				#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug, Status updates now: ")));
				HostSerial.print(showStatus ? strcpy_P(ioBuffer, PSTR("enabled")) : strcpy_P(ioBuffer, PSTR("disabled")));
				HostSerial.print(strcpy_P(ioBuffer, PSTR("; interval = ")));
				HostSerial.print(StatusInterval);
				HostSerial.println(strcpy_P(ioBuffer, PSTR("msec")));
				#endif
			} break;

			//
			//  BRAKE CONTROL
			//
			case 'b': {
				// test for proper command length
				if (len != 3) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed BRAKE command - length != 3")));
					goto finally;
				}

				// read the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(1) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed BRAKE command - Invalid port")));
					goto finally;
				}

				// read the brake status, zero (disable) or one (enable)
				uint8_t state = rxBuffer.charAt(2) - '0';
				if (state != 0 && state != 1) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed BRAKE command - Invalid state")));
					goto finally;
				}
				
				// write in the new state
				if (!motors[port]->SetBrake(state)) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Brake command unsupported by port")));
					goto finally;
				}
			} break;

			//
			//  PULSE/STEP - momentary movement
			//
			case 't':
			case 'p': {
				// need nine characters for PULSE/STEP
				if (len != 9) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed PULSE/STEP command - length != 9")));
					goto finally;
				}

				// loop counter used to copy bytes for conversion
				uint8_t i;

				// parse the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(2) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed PULSE/STEP command - Invalid port")));
					goto finally;
				}

				// parse the duration, a four-byte ASCII hex number (range is 0000-FFFF, or 0-65535 msec)
				for(i = 0; i != 4; ++i) {
					char ch = rxBuffer.charAt(i + 3);
					if (!isxdigit(ch)) {
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed PULSE/STEP command - Invalid length")));
						goto finally;
					}
					ioBuffer[i] = ch;
				}
				ioBuffer[i] = 0;
				unsigned duration = strtoul(ioBuffer, NULL, 16);

				// parse the effort, a two-byte ASCII hex number (range is 00-FF)
				for(i = 0; i != 2; ++i) {
					char ch = rxBuffer.charAt(i + 7);
					if (!isxdigit(ch)) {
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed PULSE/STEP command - Invalid effort")));
						goto finally;
					}
					ioBuffer[i] = ch;
				}
				ioBuffer[i] = 0;
				unsigned effort = strtoul(ioBuffer, NULL, 16);

				// parse the direction/polarity
				int polarity;
				switch(tolower(rxBuffer.charAt(1))) {
					case 'd':
						polarity = 0;
						break;
					case 'u':
						polarity = 1;
						break;
					default:
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed PULSE/STEP command - Invalid direction")));
						goto finally;
				}

				// command the motor
				if (cmdChar == 'p') {
					if (!motors[port]->Pulse(polarity, effort, duration)) {
						state.RecordPulse();
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, PULSE command not supported by port")));
						goto finally;
					}
				} else {
					if (!motors[port]->Step(polarity, effort, duration)) {
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, STEP command not supported by port")));
						goto finally;
					}
				}

				#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, (cmdChar == 'p') ? PSTR("#debug, PULSE Port = ") : PSTR("#debug, STEP Port = ")));
				HostSerial.print(port);
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Effort = ")));
				HostSerial.print((effort * 100) / 255);
				HostSerial.print(strcpy_P(ioBuffer, PSTR("%, Direction = ")));
				HostSerial.print(polarity ? "Forward" : "Reverse");
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Duration = ")));
				HostSerial.print(duration);
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", now = ")));
				HostSerial.print(now);
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", exp = ")));
				HostSerial.println(now + duration);
				#endif
			} break;

			//
			//  MOVE - continuous movement
			//
			case 'm': {
				// need five characters for MOVE
				if (len != 5) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed MOVE command - length != 5")));
					goto finally;
				}

				// loop counter used to copy bytes for conversion
				int i;

				// parse the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(2) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed MOVE command - Invalid port")));
					goto finally;
				}

				// parse the effort - a two-byte ASCII hex number (range is 00-FF)
				for(i = 0; i != 2; ++i) {
					char ch = rxBuffer.charAt(i + 3);
					if (!isxdigit(ch)) {
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed MOVE command - Invalid effort")));
						goto finally;
					}
					ioBuffer[i] = ch;
				}
				ioBuffer[i] = 0;
				unsigned effort = strtoul(ioBuffer, NULL, 16);

				// parse the direction/polarity
				int polarity;
				switch(tolower(rxBuffer.charAt(1))) {
					case 'd':
						polarity = 0;
						break;
					case 'u':
						polarity = 1;
						break;
					default:
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed MOVE command - Invalid direction")));
						goto finally;
				}
				
				// command the motor
				if (!motors[port]->Move(polarity, effort)) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Move command not supported by port")));
					goto finally;
				}

				#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug, MOVE Port = ")));
				HostSerial.print(port);
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Effort = ")));
				HostSerial.print((effort * 100) / 255);
				HostSerial.print(strcpy_P(ioBuffer, PSTR("%, Direction = ")));
				HostSerial.println(polarity ? strcpy_P(ioBuffer, PSTR("Forward")) : strcpy_P(ioBuffer, PSTR("Reverse")));
				#endif
			} break;
			
			#if USE_ACCELSTEPPER == true
			//
			//  RESET: zero/reset step counter
			//
			case 'r': {
				// need port for ZERO
				if (len != 2) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ZERO command - length != 2")));
					goto finally;
				}

				// parse the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(1) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ZERO command - Invalid port")));
					goto finally;
				}
				
				// call the motor
				if (!motors[port]->ZeroPosition()) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, ZERO command not supported by port")));
					goto finally;
				}
			} break;
			
			//
			//  POSITION: stepper position reading
			//
			case 'x': {
				// need nine characters for PULSE/STEP
				if (len != 2) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed POSITION command - length != 2")));
					goto finally;
				}

				// parse the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(1) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed POSITION command - Invalid port")));
					goto finally;
				}
				
				// call the motor
				long counter;
				if (!motors[port]->GetPosition(&counter)) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, POSITION command not supported by port")));
					goto finally;
				}

				// respond
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#x"))); // command
				HostSerial.print(rxBuffer.charAt(1)); // port
				HostSerial.print(',');
				HostSerial.println(counter);
			} break;

			//
			//  GOTO - seek to position
			//
			case 'g': {
				// need at least three chars
				if (len < 3) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed GOTO command - missing target")));
					goto finally;
				}

				// loop counter used to copy bytes for conversion
				uint8_t i;

				// parse the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(1) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed GOTO command - Invalid port")));
					goto finally;
				}

				// parse the target
				for (i = 0; i != rxBuffer.length() - 2; ++i) {
					ioBuffer[i] = rxBuffer.charAt(i + 2);
				}
				ioBuffer[rxBuffer.length()] = 0;
				long target = atol(ioBuffer);

				// command the motor
				if (!motors[port]->Seek(target, 255)) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, GOTO command not supported by port")));
					goto finally;
				}

				#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug, GOTO Port = ")));
				HostSerial.print(port);
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Target = ")));
				HostSerial.println(target);
				#endif
			} break;

			//
			//  ENABLE: update an enable PWM value ("e" + port + offset + hex enable + hex disable, e.g., "e01FF00")
			//
			case 'e': {
				// need seven characters for PWM
				if (len != 7 && len != 3) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ENABLE command")));
					goto finally;
				}

				// parse the output port - zero is the first port
				uint8_t port = rxBuffer.charAt(1) - '0';
				if (port < 0 || port >= NUM_PORTS) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ENABLE command - Invalid port")));
					goto finally;
				}
				
				// parse the enable pin number - zero is the first pin
				uint8_t idx = rxBuffer.charAt(2) - '0';
				if (idx < 0 || idx >= 10) {
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ENABLE command - Invalid index")));
					goto finally;
				}
				
				if (len == 3) { // GET
					// call the motor
					EnablePin *ep = motors[port]->GetEnable(idx);
					if (ep == 0) {
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, ENABLE pin unavailable")));
						goto finally;
					}

					HostSerial.print(strcpy_P(ioBuffer, PSTR("#enable,")));
					HostSerial.print((uint16_t)(port));
					HostSerial.print(strcpy_P(ioBuffer, PSTR(",")));
					HostSerial.print((uint16_t)(idx));
					HostSerial.print(strcpy_P(ioBuffer, PSTR(",")));
					HostSerial.print((uint16_t)(ep->EnablePWM));
					HostSerial.print(strcpy_P(ioBuffer, PSTR(",")));
					HostSerial.println((uint16_t)(ep->DisablePWM));
				} else if (len == 7) { // SET
					uint8_t i;
					
					// parse the ON PWM value - a two-byte ASCII hex number (range is 00-FF)
					for(i = 0; i != 2; ++i) {
						char ch = rxBuffer.charAt(i + 3);
						if (!isxdigit(ch)) {
							HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ENABLE command - Invalid ON value")));
							goto finally;
						}
						ioBuffer[i] = ch;
					}
					ioBuffer[i] = 0;
					uint8_t enable = strtoul(ioBuffer, NULL, 16);
					
					// parse the OFF PWM value - a two-byte ASCII hex number (range is 00-FF)
					for(i = 0; i != 2; ++i) {
						char ch = rxBuffer.charAt(i + 5);
						if (!isxdigit(ch)) {
							HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Malformed ENABLE command - Invalid OFF value")));
							goto finally;
						}
						ioBuffer[i] = ch;
					}
					ioBuffer[i] = 0;
					uint8_t disable = strtoul(ioBuffer, NULL, 16);
					
					// call the motor
					EnablePin *ep = motors[port]->GetEnable(idx);
					if (ep == 0) {
						HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, ENABLE pin unavailable")));
						goto finally;
					}
					ep->EnablePWM = enable;
					ep->DisablePWM = disable;
					motors[port]->Stop(); // stop the motor AND commit the new values
				}
			} break;
			#endif
			#if PERSIST_SETTINGS == true
			//
			//  READ: get startup settings from EEPROM
			//
			case 'a': {
				// restore PWM setting to EEPROM
				load_eeprom();
			} break;

			//
			//  WRITE: commit settings to EEPROM
			//
			case 'w': {
				// store PWM setting to EEPROM
				save_eeprom();
			} break;

			//
			//  FACTORY: reset EEPROM data
			//
			case 'f': {
				// store PWM setting to EEPROM
				clear_eeprom();
			} break;
			#endif

			//
			//  everything else...
			//
			default: {
				HostSerial.println(strcpy_P(ioBuffer, PSTR("#error, Unknown command received")));
				goto finally;
			} break;
		}

		// echo response to the host if successfully read and executed
		HostSerial.println("#OK," + rxBuffer); 

finally:
		// next command, please
		rxBuffer = "";
		cmdReady = false;
	}

	// service the serial port
	while (HostSerial.available()) {
		// get the new byte
		char inChar = (char)HostSerial.read(); 
		// if the incoming character is a newline, set a flag
		// so the main loop can do something about it
		if (inChar == Delimiter) {
			#ifdef ECHO
			HostSerial.print("\r\n");
			#endif
			cmdReady = true;
			return;
		} 

		// ignore whitespace and punctuation, but allow the
		//    host to supply them if it wants to make the
		//    command more clear
		if (isSpace(inChar) || (isPunct(inChar) && (inChar != '-') && (inChar != '+'))) continue;

		#ifdef ECHO
		HostSerial.print(inChar);
		#endif
		// add other bytes to the rxBuffer
		if (rxBuffer.length() < (MaxLine - 1))
			rxBuffer += inChar;
	}
}

// EOF: AutoCap_ArduinoUno_v1.ino
