/*
 *
 *
 *    EnablePin.h
 *
 *    Enable pin definition.
 *
 *    AutoCap Arduino Firmware
 *    Version 1.5
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#ifndef AUTOCAP_ENABLEPIN_H
#define AUTOCAP_ENABLEPIN_H

// Arduino environment stuff
#include <Arduino.h>

//
//  struct EnablePin - definition for an ENABLE output pin
//
struct EnablePin {
	const uint8_t Pin;		// Optional enable/PWM pin
	const bool IsPWM;		// indicates Enable pin is PWM 
	uint8_t EnablePWM;		// the PWM value for Enable pin when ON
	uint8_t DisablePWM;		// the PWM value for Enable pin when OFF
	bool DisableOnStop;		// indicates Enable pin should be de-asserted on Stop()

	//
	//  Constructor Arguments:
	//
	//     pinEnable     - GPIO/PWM output for ENABLE pin; used to power up motor
	//                     when running, and power-down when stopped.
	//     enableIsPwm   - set to true if the ENABLE pin is a PWM-capable pin, and
	//                     PWM output on that pin is desired; when set to false,
	//                     the ENABLE pin is just a boolean (GPIO) output.
	//     pwmValueOn    - the value of the ENABLE pin when the motor is running;
	//                     when enableIsPwm == false, this should be either HIGH
	//                     or LOW; when enableIsPwm == true, this is the PWM duty
	//                     cycle to write to the ENABLE pin (0 = 0%, 255 = 100%).
	//     disableOnStop - set to true if the ENABLE pin output value should be
	//                     different when the motor is stopped; false if the pin's
	//                     value is constant.
	//     pwmValueOff   - the value of the ENABLE pin when the motor is stopped;
	//                     when enableIsPwm == false, this should be either HIGH
	//                     or LOW; when enableIsPwm == true, this is the PWM duty
	//                     cycle to write to the ENABLE pin (0 = 0%, 255 = 100%).
	//
	EnablePin(
			uint8_t pin,				// ENABLE/PWM pin
			bool isPwm = false,			// ENABLE line is PWM
			uint8_t enablePwm = 0xFF,	// PWM ENABLE value (0-255 or 0x00-0xFF)
			bool disableOnStop = false,	// disable-on-stop flag
			uint8_t disablePwm = 0x00	// PWM DISABLE value (0-255 or 0x00-0xFF)
			) : Pin(pin), IsPWM(isPwm), EnablePWM(enablePwm),
				DisablePWM(disablePwm), DisableOnStop(disableOnStop) {
		// nop
	}
};
#endif
