/*
 *
 *
 *    Interfaces.h
 *
 *    Motor interface.
 *
 *    AutoCap Arduino Firmware
 *    Version 1.5
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#ifndef AUTOCAP_INTERFACES_H
#define AUTOCAP_INTERFACES_H

#include <Arduino.h>
#include "EnablePin.h"

//
//  Generic Motor Interface
//
class IMotor {
	public:
		virtual ~IMotor() { /* nop */ }

		//
		//  Setup() - initialize the hardware
		//
		virtual void Setup() = 0;

		//
		//  Run() - update state, timers, actions, etc.; this should be called once per loop() iteration
		//
		virtual void Run() = 0;
		
		//
		//  Stop() - stop the motor, regardless of what it is doing
		//
		virtual bool Stop() = 0;
		
		//
		//  Pulse(...) - pulse the motor with a specific direction, speed, and duration
		//
		virtual bool Pulse(bool direction, uint8_t speed, uint16_t howLong) = 0;
		
		//
		//  Step(...) - move step-equipped motor specific number of steps at given rate
		//
		virtual bool Step(bool direction, uint8_t speed, uint16_t steps) = 0;
		
		//
		//  Move(...) - move the motor continuously with a specific direction and speed
		//
		virtual bool Move(bool direction, uint8_t speed) = 0;
		
		//
		//  SetBrake(state) - set state of braking feature
		//
		virtual bool SetBrake(bool state) = 0;
		
		//
		//  SetBrake(state) - set state of braking feature
		//
		virtual bool GetBrake() = 0;
		
		//
		//  LED(state) - update the 'port selected' LED state (or other similar indicator)
		//
		virtual void LED(bool state) = 0;

        //
		//  ReadCurrent() - read the present current (I) value
		//
		virtual float ReadCurrent() = 0;
		
		//
		//  Read the current stepper position; return true if supported
		//
		virtual bool GetPosition(long *outValue) = 0;
		
		//
		//  Reset the current stepper position to zero; return true if supported
		//
		virtual bool ZeroPosition(void) = 0;
		
		//
		//  Move to a specific position
		//
		virtual bool Seek(long target, uint8_t speed) = 0;
		
		//
		//  Get an enable pin definition
		//
		virtual EnablePin *GetEnable(uint8_t idx) = 0;

		//
		//  Returns true iff the motor is moving
		//
		virtual bool IsRunning() const = 0;
};

#endif // AUTOCAP_INTERFACES_H
