/*
 *
 *
 *    persist.h
 *
 *    AutoCap Arduino Firmware
 *    Version 1.5
 *
 *    User settings persistence support (EEPROM support).
 *
 *    Note: this file is a simple refactoring of the main INO file, and
 *          cannot be built outside of a #include from that file.
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#ifndef __AUTOCAP_PERSIST_H
#define __AUTOCAP_PERSIST_H

#include "EEPROM.h"

// define the unique string for EEPROM detection
#define EEPROM_KEY ("AutoCap01.50")

//
//  Binary DTO for Brake Settings
//
struct BrakeBlock {
	uint8_t port;    // port number
	bool    brake;   // brake status

	BrakeBlock() { port = brake = 0; }
};

//
//  Binary DTO for Enable Settings
//
struct EnableBlock {
	uint8_t port;    // port number
	uint8_t index;   // enable pin index
	uint8_t enable;  // enable PWM value (steppers only)
	uint8_t disable; // disable PWM value (steppers only)

	EnableBlock() { port = index = enable = disable = 0; }
};

//
//  Binary DTO for State fields
//
struct StateBlock {
	InputModes InputMode;
	uint16_t Width;
	uint16_t Steps;
	uint8_t UserPWM;
	uint8_t UserPort;
};

//
//  write object to EEPROM
//
inline int write_eeprom(int pos, void *obj, int size) {
	char *cp = (char*)(obj);
	for (uint8_t i = 0; i != size; ++i) {
		EEPROM.write(pos + i, *cp++);
	}
	return pos + size;
}

//
//  read object from EEPROM
//
inline int read_eeprom(int pos, void *obj, int size) {
	char *cp = (char*)(obj);
	for (uint8_t i = 0; i != size; ++i) {
		*cp++ = EEPROM.read(pos + i);
	}
	return pos + size;
}

//
//  clear the EEPROM
//
inline void clear_eeprom() {
	for(uint16_t i = 0; i != E2END + 1; ++i) {
		EEPROM.write(i, 0);
	}
}

//
//  update just the state vars in the EEPROM
//
inline void update_state_in_eeprom() {
	// EEPROM pointer
	uint16_t pos = 0;

	// find and read through the header
	{
		char *header = strcpy_P(ioBuffer, PSTR(EEPROM_KEY));
		uint8_t count = strlen(header);
		for (uint8_t i = 0; i != count; ++i) {
			if (EEPROM.read(i) != ioBuffer[i]) {
#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,header mismatch @ ")));
				HostSerial.print((int)(i));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" contents: ")));
				for (uint8_t j = 0; j != count; ++j) {
					HostSerial.print((char)(EEPROM.read(j)));
				}
				HostSerial.println();
#endif
				// can't find the config block, give up
				return;
			}
		}
		pos = count;
	}
	// write state block over the top of the existing one
	{
		// copy
		StateBlock sb;
		sb.InputMode = state.InputMode;
		sb.Width = state.Width;
		sb.Steps = state.Steps;
		sb.UserPWM = state.UserPWM;
		sb.UserPort = state.UserPort;

		// write
		pos = write_eeprom(pos, &sb, sizeof(StateBlock));
	}
}

//
//  store EEPROM state
//
inline void save_eeprom() {
	// EEPROM pointer
	uint16_t pos = 0;

	// write out header
	{
		strcpy_P(ioBuffer, PSTR(EEPROM_KEY));
		uint8_t count = strlen(ioBuffer);
		pos = write_eeprom(pos, ioBuffer, count);
#ifdef DEBUG
		HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,header key length ")));
		HostSerial.print((int)(count));
		HostSerial.println();
#endif
	}

	// write state block
	{
		// copy
		StateBlock sb;
		sb.InputMode = state.InputMode;
		sb.Width = state.Width;
		sb.Steps = state.Steps;
		sb.UserPWM = state.UserPWM;
		sb.UserPort = state.UserPort;

		// write
		pos = write_eeprom(pos, &sb, sizeof(StateBlock));
	}
	// write brake blocks
	{
#ifdef DEBUG
		HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,write brake count @ ")));
		HostSerial.print((int)(pos));
		HostSerial.println();
#endif
		// write the number of brake blocks
		uint16_t count = NUM_PORTS;
		pos = write_eeprom(pos, &count, sizeof(count));

		// write out the blocks
		BrakeBlock bb;
		for (uint8_t port = 0; port != NUM_PORTS; ++port) {
			bb.port = port;
			bb.brake = motors[port]->GetBrake();
#ifdef DEBUG
			HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,write brake block ")));
			HostSerial.print((int)(port));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" @ ")));
			HostSerial.print((int)(pos));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" max ")));
			HostSerial.print((int)(count));
			HostSerial.println();
#endif
			// write the block
			pos = write_eeprom(pos, &bb, sizeof(BrakeBlock));
		}
	}
	// write enable blocks
	{
		// count the number of enable pins
		uint16_t count = 0;
		for (uint8_t port = 0; port != NUM_PORTS; ++port) {
			for(uint8_t pin = 0; pin != 0xFF; ++pin) {
				EnablePin *ep = motors[port]->GetEnable(pin);
				if (!ep) {
					break;
				}
				++count;
			}
		}
#ifdef DEBUG
		HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,write enable count @ ")));
		HostSerial.print((int)(pos));
		HostSerial.print(strcpy_P(ioBuffer, PSTR(" count ")));
		HostSerial.print((int)(count));
		HostSerial.println();
#endif
		// write the number of enable pins
		pos = write_eeprom(pos, &count, sizeof(count));

		// write out the blocks
		EnableBlock eb;
		for (uint8_t port = 0; port != count; ++port) {
			for(uint8_t pin = 0; pin != 0xFF; ++pin) {
				EnablePin *ep = motors[port]->GetEnable(pin);
				if (!ep) {
					break;
				}

				eb.port = port;
				eb.index = pin;
				eb.enable = ep->EnablePWM;
				eb.disable = ep->DisablePWM;

				pos = write_eeprom(pos, &eb, sizeof(EnableBlock));
#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,write enable block ")));
				HostSerial.print((int)(port));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" @ ")));
				HostSerial.print((int)(pos));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" idx ")));
				HostSerial.print((int)(pin));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" max ")));
				HostSerial.print((int)(count));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" en ")));
				HostSerial.print((int)(eb.enable));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" dis ")));
				HostSerial.print((int)(eb.disable));
				HostSerial.println();
#endif
			}
		}
	}
}

//
//  restore EEPROM state
//
inline void load_eeprom(bool quiet = false) {
	uint16_t pos = 0;
	// read the header
	{
		char *header = strcpy_P(ioBuffer, PSTR(EEPROM_KEY));
		uint8_t count = strlen(header);
		for (uint8_t i = 0; i != count; ++i) {
			if (EEPROM.read(i) != ioBuffer[i]) {
#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,header mismatch @ ")));
				HostSerial.print((int)(i));
				HostSerial.print(strcpy_P(ioBuffer, PSTR(" contents: ")));
				for (uint8_t j = 0; j != count; ++j) {
					HostSerial.print((char)(EEPROM.read(j)));
				}
				HostSerial.println();
#endif
				// can't find the config block, give up
				if (!quiet)
					HostSerial.println(strcpy_P(ioBuffer, PSTR("#error,EEPROM header not found")));
				return;
			}
		}
		pos = count;
	}
	// read the state block
	{
		StateBlock sb;
		pos = read_eeprom(pos, &sb, sizeof(StateBlock));
		state.InputMode = sb.InputMode;
		state.Width = sb.Width;
		state.Steps = sb.Steps;
		state.UserPWM = sb.UserPWM;
		state.UserPort = sb.UserPort;
#ifdef DEBUG
		HostSerial.print("#debug,Read state ");
		HostSerial.print(state.InputMode);
		HostSerial.print(",");
		HostSerial.print(state.Width);
		HostSerial.print(",");
		HostSerial.print(state.Steps);
		HostSerial.print(",");
		HostSerial.print(state.UserPWM);
		HostSerial.print(",");
		HostSerial.print(state.UserPort);
		HostSerial.println();
#endif
	}
	// read the brake blocks
	{
		// count the number of brake blocks
		uint16_t count = 0;
		pos = read_eeprom(pos, &count, sizeof(count));

		// read the blocks
		BrakeBlock bb;
		for (uint8_t port = 0; port != count; ++port) {
#ifdef DEBUG
			HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,read brake block ")));
			HostSerial.print((int)(port));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" @ ")));
			HostSerial.print((int)(pos));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" max ")));
			HostSerial.print((int)(count));
			HostSerial.println();
#endif
			// read block
			pos = read_eeprom(pos, &bb, sizeof(BrakeBlock));

			if (bb.port >= NUM_PORTS) {
				// too many ports in configuration, bail out
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#error,EEPROM has invalid brake block: ")));
				HostSerial.print((int)(bb.port));
				HostSerial.println();
				return;
			}

			// commit setting
			motors[bb.port]->SetBrake(bb.brake);
		}
	}
	{
		// count the number of enable pins
		uint16_t count = 0;
		EnableBlock eb;
		pos = read_eeprom(pos, &count, sizeof(count));
		for (uint8_t port = 0; port != count; ++port) {
			// read block
			pos = read_eeprom(pos, &eb, sizeof(EnableBlock));

			if (eb.port >= NUM_PORTS) {
				// too many ports in configuration, bail out
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#error,EEPROM has invalid enable block: ")));
				HostSerial.print((int)(eb.port));
				HostSerial.println();
				return;
			}

			// commit settings
			EnablePin *ep = motors[eb.port]->GetEnable(eb.index);
			if (!ep) {
				// bad pin offset; give up
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#error,EEPROM has invalid pin block: ")));
				HostSerial.print((int)(eb.index));
				HostSerial.println();
				return;
			}
			ep->EnablePWM = eb.enable;
			ep->DisablePWM = eb.disable;
#ifdef DEBUG
			HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,read enable block ")));
			HostSerial.print((int)(port));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" @ ")));
			HostSerial.print((int)(pos));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" max ")));
			HostSerial.print((int)(count));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" idx ")));
			HostSerial.print((int)(eb.index));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" en ")));
			HostSerial.print((int)(eb.enable));
			HostSerial.print(strcpy_P(ioBuffer, PSTR(" dis ")));
			HostSerial.print((int)(eb.disable));
			HostSerial.println();
#endif
		}

		// commit the new PWM settings
		for (int i = 0; i != NUM_PORTS; ++i) {
			motors[i]->Stop();
		}
	}
}

#endif
